#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>
#include <png.h>
#include <setjmp.h>
#include <sys/stat.h>

// #define USEICONV
#ifdef USEICONV
#include <iconv.h>
#else
typedef int iconv_t;
#endif

iconv_t iso_to_utf8, utf16_to_utf8;

int embedpng = 0;
int allinone = 0;
char *filebase;

void die(const char s[], ...)
{
  va_list args;

  fflush(stdout);
  fflush(stderr);

  va_start(args, s);
  vfprintf(stderr, s, args);
  va_end(args);

  fputs("\nPress any key to abort.", stderr);

  getch();
  fputs("\n", stderr);

  exit(1);
}

void *my_alloc(size_t len, const char *name)
{
  void *p;

  p = malloc(len);
  if (!p) die("Not enough memory for %s, wanted %d", name, len);

  return p;
}

void *my_realloc(void *p, size_t len, const char *name)
{
  p = realloc(p, len);
  if (!p) die("Not enough memory for %s, wanted %d", name, len);

  return p;
}

void putxmlstr(FILE *f, const char *str, iconv_t conv, uint32_t minlen)
{
  size_t i, buflen, irrev, inleft, outleft;
  const char *inbuf;
  char *buffer, *outbuf;

#ifndef USEICONV
  conv = 0;
#endif

  if (!minlen) minlen = strlen(str);

  buflen = minlen*5;
  buffer = my_alloc(buflen, "buffer");

  if (conv) {
#ifdef USEICONV
  inleft = minlen;
  outleft = buflen;
  inbuf = str;
  outbuf = buffer;
  irrev = iconv(conv, &inbuf, &inleft, &outbuf, &outleft);
  if (irrev == -1) die("Error at %s while converting %s to utf8: %s", inbuf, str, strerror(errno));
  if (irrev) die("Found %d irreversible conversions in %s", irrev, str);

  buflen -= outleft;
#endif
  } else {
  memcpy(buffer, str, minlen);
  buflen = minlen;
  }

  for (i=0; i<buflen; i++)
	switch (buffer[i]) {
		case '"': fputs("&quot;", f); break;
		case '<': fputs("&lt;", f); break;
		case '>': fputs("&gt;", f); break;
		case '&': fputs("&amp;", f); break;
//		case 0 ... 31: fprintf(f, "&#x%02X;", buffer[i]); break;
		default: fputc(buffer[i], f); break;
	}
  free(buffer);
}

struct varinf {
	int ofs;
	int size;
	int num;
	const char *name;
	struct varinf *structvars;
	const char **flags;
};

char indent[11] = "\t\t\t\t\t\t\t\t\t\t";
#define INDENT (indent+10-indentation)

uint32_t getvalue(uint8_t *data, int size)
{
  uint32_t value = *(uint32_t*)(data);
  value &= 0xffffffff >> (32-8*size);

  return value;
}

int32_t getsvalue(uint8_t *data, int size)
{
  int32_t value = *(int32_t*)(data);
  value = (value << (32-8*size)) >> (32-8*size);

  return value;
}


uint32_t dumpflags(FILE *xml, uint8_t *data, const char **flags, int size, int indentation)
{
  uint32_t value, i, state;
  const char *name;
  char defname[32];

  if (size > 4) die("Can't dump flags with %d bytes", size);

  value = getvalue(data, size);

  for (i=0; i<size*8; i++) {
	if (*flags) {
		name = *flags;
		flags++;
	} else
		name = "";

	if (!strlen(name)) {
		sprintf(defname, "bit%d", i);
		name = defname;
	}
	state = (value & (1 << i)) >> i;
	if ((name != defname) || state)
		fprintf(xml, "%s<bit name=\"%s\">%d</bit>\n", INDENT, name, state);
  }
  return size;
}

uint32_t dodump(FILE *xml, uint8_t *data, struct varinf *vars, int indentation)
{
  int i, j;
  uint32_t dumped, totaldumped, ofs;
  const char *name, *fname;
  char defname[32], arrname[32];

  totaldumped = 0;
  ofs = 0;

  for (i=0; vars[i].ofs != -1; i++) {
	fname = vars[i].name;
	if (!strlen(fname)) {
		sprintf(defname, "field_%X", vars[i].ofs);
		fname=defname;
	}
	if (ofs != vars[i].ofs) die("Structure is invalid at %X", ofs);
	for (j=0; j<vars[i].num; j++) {
		name = fname;
		if (vars[i].num > 1) {
			sprintf(arrname, "%s[%d]", fname, j);
			name = arrname;
		}

		if (vars[i].flags) {
			fprintf(xml, "%s<bitmask name=\"%s\" size=\"%d\">\n", INDENT, name, vars[i].size);
			dumped = dumpflags(xml, data+ofs, vars[i].flags, vars[i].size, indentation+1);
			fprintf(xml, "%s</bitmask>\n", INDENT);
		} else if (vars[i].structvars) {
			fprintf(xml, "%s<structure name=\"%s\">\n", INDENT, name);
			dumped = dodump(xml, data+ofs, vars[i].structvars,
				indentation+1);

			fprintf(xml, "%s</structure>\n", INDENT);
		} else if (strlen(vars[i].name)) {
			fprintf(xml, "%s<variable name=\"%s\" size=\"%X\">%d</variable>\n", INDENT, name, vars[i].size, getvalue(data+ofs, vars[i].size));
			dumped = vars[i].size;
		} else {
			uint32_t value = getvalue(data+ofs, vars[i].size);
			if (value)
				fprintf(xml, "%s<unknown name=\"%s\" size=\"%X\">%d</unknown>\n", INDENT, name, vars[i].size, value);
			dumped = vars[i].size;
		}
		ofs += dumped;
		totaldumped += dumped;
	}
  }
  return totaldumped;
}

uint32_t dumpvars(FILE *xml, uint8_t *data, struct varinf *vars)
{
  return dodump(xml, data, vars, 1);
}

uint32_t dumplang(FILE *xml, uint8_t *data)
{
  int language;
  uint32_t ofs;

  ofs = 0;
  while ( (language = data[ofs++]) != 0xff ) {
	fprintf(xml, "\t<description language=\"%d\">", language);
	putxmlstr(xml, data+ofs, language<9?iso_to_utf8:utf16_to_utf8, 0);
	fputs("</description>\n", xml);
	ofs += strlen(data+ofs) + 1;
  }

  return ofs;
}

uint32_t dumpuseobj(FILE *xml, uint8_t *data, int num, int total, const char *type, int class)
{
  char arrtype[32];

  if (total > 1) {
	sprintf(arrtype, "%s[%d]", type, num);
	type = arrtype;
  }

  if (class != data[0]) die("Not correct class for %s; want %d got %d", type, class, data[0]);

  fprintf(xml, "\t<useobject desc=\"%s\" class=\"%d\">%.8s</useobject>\n",
	type, class, data+4);

  return 16;
}

uint32_t dumpcap(FILE *xml, uint8_t *data, int num, int total)
{
  uint32_t dumped;
  int capacity, cargotype, refcap;

  if (total);

  capacity = data[0];

  dumped = 1;

  if (capacity) {
	fprintf(xml, "\t<cargo num=\"%d\" capacity=\"%d\">\n", num, capacity);
	while (getvalue(data+dumped, 2) != 0xffff) {
		cargotype = data[dumped++];

		refcap = getvalue(data+dumped, 2);
		dumped += 2;

		fprintf(xml, "\t\t<cargotype id=\"%d\">%d</cargotype>\n", cargotype, refcap);
	}
	dumped += 2;
	fputs("\t</cargo>\n", xml);
  }

  return dumped;
}

uint32_t putspriterow(uint8_t *data, int flags, int y, int width, png_byte *row)
{
  uint32_t offset, last, ofs, len;

  memset(row, 0, width);

  if (flags & 4) {
	offset=getvalue(data+2*y, 2);

	do {
		last=data[offset]   & 0x80;
		len=data[offset++] & 0x7f;
		ofs=data[offset++];

		memcpy(row+ofs, data+offset, len);

		offset+=len;
	} while (!last);
  } else {
	memcpy(row, data+y*width, width);
	offset = (y+1)*width;
  }
  return offset;
}

void xml_write(png_structp png_ptr, png_bytep data, png_size_t size)
{
//  putxmlstr(png_ptr->io_ptr, data, iso_to_utf8, size);
  fwrite(data, 1, size, png_ptr->io_ptr);
}

void xml_flush(png_structp png_ptr)
{
  fflush(png_ptr->io_ptr);
}

#include "lomopal.h"

uint32_t makepng(FILE *xml, FILE *png, uint8_t *data, uint32_t width, uint32_t height, uint32_t flags)
{
   png_structp png_ptr;
   png_infop info_ptr;
   png_colorp palette;
   png_byte *row;

   int i,j;
   uint32_t dumped;

   dumped = 0;

#if !defined(__BORLANDC__) || __BORLANDC__ > 0x500

//   fputs("<![CDATA[", xml);

   if (setmode(fileno(xml), O_BINARY) == 1) die("Can't switch to binary file mode");

	// create the png structures
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   if (!png_ptr) die("Error creating png write struct");

   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
	png_destroy_write_struct(&png_ptr, png_infopp_NULL);
	die("Error creating png info struct");
   }

	// error handlingg
   if (setjmp(png_jmpbuf(png_ptr))) {
		// if we get here, we had a problem writing the file
	png_destroy_write_struct(&png_ptr, &info_ptr);
	die("Error while writing png data");
   }

   if (embedpng) {
		// use our own write functions
	png_set_write_fn(png_ptr, xml, xml_write, xml_flush);
   } else if (allinone) {
   } else {
	png_init_io(png_ptr, png);
   }

	// set image dimensions
   png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_PALETTE,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// set the palette
   palette = (png_colorp)png_malloc(png_ptr, PNG_MAX_PALETTE_LENGTH * sizeof (png_color));
   if (!palette) die("Not enough memory for %s, wanted %d", "palette", PNG_MAX_PALETTE_LENGTH * sizeof (png_color));

   for (i=0; i<256; i++) {
	palette[i].red = lomopalette[i][0];
	palette[i].green = lomopalette[i][1];
	palette[i].blue = lomopalette[i][2];
   }

   png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);

	// set the gamma
   png_set_gAMA(png_ptr, info_ptr, 1.4);

	// write the png header
   png_write_info(png_ptr, info_ptr);

   if (embedpng || !allinone) {

   row = my_alloc(width, "row");

	// write the image line by line
   for (i=0; i<height; i++) {
	uint32_t offset = putspriterow(data, flags, i, width, row);

	if (offset > dumped) dumped=offset;

	png_write_row(png_ptr, row);
   }

   free(row);

   }

	// done writing, clean up
   png_write_end(png_ptr, info_ptr);
   png_free(png_ptr, palette);
   png_destroy_write_struct(&png_ptr, &info_ptr);

   if (setmode(fileno(xml), O_TEXT) == 1) die("Can't switch to text file mode");
//   fputs("]]>", xml);

#endif
   return dumped;
}

const char *spriteflags[] = {
	"hasdata", "", "chunked",
	NULL
};

uint32_t dumpsprites(FILE *xml, uint8_t *data)
{
  uint32_t i, j, num, size, totalsize, dumped;
  uint8_t *spritedata;
  FILE *png;

  num = getvalue(data, 4);
  size = getvalue(data+4, 4);
  dumped = 8;

  spritedata = data+8+16*num;
  totalsize = 0;

  for (i=0; i<num; i++) {
	uint32_t ofs,width,height,flags, spritesize;
	int32_t xofs,yofs;

	ofs=getvalue(data+dumped,4);  dumped+=4;
	width=getvalue(data+dumped,2);  dumped+=2;
	height=getvalue(data+dumped,2);  dumped+=2;
	xofs=getsvalue(data+dumped,2);  dumped+=2;
	yofs=getsvalue(data+dumped,2);  dumped+=2;
	flags=getvalue(data+dumped,4);  dumped+=4;

	fprintf(xml, "\t<sprite id=\"%d\" xofs=\"%d\" yofs=\"%d\">\n", i, xofs, yofs);
	dumpflags(xml, (uint8_t*) &flags, spriteflags, 4, 2);

	if (embedpng) {
		fputs("\t\t<pngdata>", xml);
		png = NULL;
	} else if (allinone) {
		// fprintf(xml, "\t\t<png x= y= width= height=>%s</png>\n", );
	} else {
		char *pngname;

		pngname = my_alloc(strlen(filebase)+16, "pngname");
		sprintf(pngname, "%d", num);
		j = strlen(pngname);
		strcpy(pngname, filebase);
		sprintf(pngname+strlen(pngname), "_%0*d.png", j, i);
		png = fopen(pngname, "wb");
		if (!png) die("Can't write %s: %s", pngname, strerror(errno));

		fprintf(xml, "\t\t<pngfile>%s</pngfile>\n", pngname);

		free(pngname);
	}

	spritesize = makepng(xml, png, spritedata + ofs, width, height, flags);

	if (embedpng) {
		fputs("</pngdata>\n", xml);
	} else if (!allinone) {
		fclose(png);
	}

	totalsize += spritesize;

	fputs("\t</sprite>\n", xml);
  }

  dumped += totalsize;
  if (totalsize != size) die("Sprite data different from size; wanted %d had %d", size, totalsize);

  return dumped;
}

const char *vehspriteflags[] = {
	"hassprites",
	NULL
};

struct varinf vehsprites[] = {
	{ 0x00, 1, 1, "numdir" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "" },
	{ 0x07, 1, 1, "flags", NULL, vehspriteflags },
	{ 0x08, 1, 1, "" },
	{ 0x09, 1, 1, "" },
	{ 0x0A, 1, 1, "" },
	{ 0x0B, 1, 1, "" },
	{ 0x0C, 1, 1, "" },
	{ 0x0D, 1, 1, "" },
	{ 0x0E, 1, 1, "spritenum" },
	{ 0x0F, 1, 3, "" },
	{ 0x12, 4, 1, "" },
	{ 0x16, 4, 1, "" },
	{ 0x1A, 4, 1, "" },
	{ -1 },
};

struct varinf vehunk[] = {
	{ 0x00, 1, 1, "" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "spriteind" },
	{ 0x05, 1, 1, "" },
	{ -1 },
};

const char *vehflags[] = {
	"", "", "", "",
	"", "", "rackrail", "",
	"", "anytrack", "", "",
	"dualhead", "", "refittable", "noannounce",
	NULL
};

struct varinf vehvars[] = {
	{ 0x000, 1, 2, "" },
	{ 0x002, 1, 1, "class" },
	{ 0x003, 1, 1, "" },
	{ 0x004, 1, 1, "" },
	{ 0x005, 1, 1, "" },
	{ 0x006, 1, 1, "nummods" },
	{ 0x007, 1, 1, "costind" },
	{ 0x008, 2, 1, "costfact" },
	{ 0x00A, 1, 1, "reliability" },
	{ 0x00B, 1, 1, "runcostind" },
	{ 0x00C, 2, 1, "runcostfact" },
	{ 0x00E, 1, 1, "" },
	{ 0x00F, 1, 1, "numcompat" },
	{ 0x010, 1,20, "" },
	{ 0x024, 0, 4, "", vehunk },
	{ 0x03C, 0, 4, "sprites", vehsprites },
	{ 0x0B4, 1,36, "" },
	{ 0x0D8, 2, 1, "power" },
	{ 0x0DA, 2, 1, "speed" },
	{ 0x0DC, 2, 1, "rackspeed" },
	{ 0x0DE, 2, 1, "weight" },
	{ 0x0E0, 2, 1, "flags", NULL, vehflags },
	{ 0x0E2, 1,44, "" },
	{ 0x10E, 1, 1, "" },
	{ 0x10F, 1, 1, "visfxtype" },
	{ 0x110, 1, 1, "" },
	{ 0x111, 1, 1, "" },
	{ 0x112, 1, 1, "wakefxtype" },
	{ 0x113, 1, 1, "" },
	{ 0x114, 2, 1, "designed" },
	{ 0x116, 2, 1, "obsolete" },
	{ 0x118, 1, 1, "" },
	{ 0x119, 1, 1, "startsndtype" },
	{ 0x11A, 1,64, "" },
	{ 0x15A, 1, 1, "numsnd" },
	{ 0x15B, 1, 3, "" },
	{ -1 },
};


uint32_t dumpveh(FILE *xml, uint8_t *data)
{
  int i;
  uint32_t dumped;

  dumped = dumpvars(xml, data, vehvars);
  dumped += dumplang(xml, data+dumped);
  if ((data[2] < 2) && !(getvalue(data+0xE0, 2) & 0x200))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, "tracktype", 0x11);
  for (i=0; i<data[6]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[6], "trackmod", 0x10);
  for (i=0; i<2; i++)
	dumped += dumpcap(xml, data+dumped, i, 2);
  if (data[0x10F])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, "visualeffect", 0x03);
  if (data[0x112])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, "wakeeffect", 0x3);
  if ((data[2] < 2) && (data[0xE0]&0x40))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, "rackrail", 0x10);
  for (i=0; i<data[15]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[15], "compatible", 0x17);
  if (data[0x119])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, "startsnd", 0x01);
  for (i=0; i<(data[0x15A] & 0x7f); i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x15A] & 0x7f, "soundeffect", 0x01);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

int dumpchunk(FILE *xml, uint8_t *data, int class)
{
  uint32_t dumped;

  switch (class) {
	case 0x17:
		dumped = dumpveh(xml, data);
		break;
	default:
		die("Can't translate class %02X yet", class);
  }

  return dumped;
}

uint8_t *rledecode(uint8_t *data, uint32_t len)
{
  uint8_t *chunk;
  uint32_t chunklen, chunkofs;

  chunklen = len*2;
  chunk = my_alloc(len*2, "chunk");

  chunkofs = 0;
  while (len>0) {
	int8_t rle;
	int run;

	rle = *(int8_t*) data++; len--;
	run = abs(rle)+1;

	if (rle<0) {
		int i;
		uint8_t byte = *data++; len--;

		memset(chunk+chunkofs, byte, run);
	} else {
		memcpy(chunk+chunkofs, data, run);
		len-=run;
		data+=run;
	}
	chunkofs += run;
	while (chunkofs > chunklen) {
		chunklen <<= 1;
		chunk = my_realloc(chunk, chunklen, "chunk");
	}
  }

  chunk = my_realloc(chunk, chunkofs, "chunk");

  return chunk;
}

int rleencode(uint8_t *data, uint32_t len, uint8_t **rleout, uint32_t *rlelen)
{
  uint8_t *out;
  uint32_t start, ofs, outlen, outofs;
  int run;
  uint8_t byte, rle;

  outlen = len*2;
  out = my_alloc(outlen, "out");

  ofs = 0;
  outofs = 0;
  while (len) {
	start = ofs;

	byte = data[ofs++];
	len--;

	run = 1;
	if (len && (byte == data[ofs])) {
		while (len && (byte == data[ofs]) && (run < 127)) {
			run++;
			ofs++;
			len--;
		}
		rle = -(run-1);
		out[outofs++] = rle;
		out[outofs++] = byte;
	} else {
		while (len && (byte != data[ofs]) && (run < 126)) {
			byte = data[ofs];
			run++;
			ofs++;
			len--;
		}
		if (len) {
			run--;
			ofs--;
			len++;
		}
		rle = run-1;
		out[outofs++] = rle;
		memcpy(out+outofs, data+start, run);
		outofs += run;
	}
  }
  *rlelen = outofs;
  *rleout = out;
  return 1;
}

// define a bit rotate (gcc optimizes this into an i386 ROL instruction)
#define _rotl32(a,n) (((a)<<(n))|((a)>>(32-n)))

uint32_t checksum(uint8_t *data, uint32_t len)
{
  uint32_t i, c;

  c = 0xF369A75B;

  c = _rotl32(c ^ data[0], 11);
  for (i=4; i<12; i++)
	c = _rotl32(c ^ data[i], 11);
  for (i=16; i<len; i++)
	c = _rotl32(c ^ data[i], 11);

  return c;
}

int decode(char *datname)
{
  int class;
  uint32_t subclass;
  uint8_t *chunk;
  uint32_t dumped;

  uint8_t *data, *end;
  struct stat st;
  FILE *dat, *xml;
  char *xmlname;

  xmlname = my_alloc(strlen(filebase)+5, "xmlname");
  strcpy(xmlname, filebase);
  strcat(xmlname, ".xml");

  dat = fopen(datname, "rb");
  if (!dat) die("Can't read %s: %s", datname, strerror(errno));

  if (fstat(fileno(dat), &st)) die("Can't stat %s: %s", datname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  xml = fopen(xmlname, "wt");
  if (!xml) die("Can't write %s: %s", xmlname, strerror(errno));
  free(xmlname);

  data = my_alloc(st.st_size, "data");

  fread(data, 1, st.st_size, dat);

  if (data[3] != 0x11) die("File %s is not a .dat file", datname);
  end = data+st.st_size;

  fputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", xml);

  class = data[0];
  subclass = getvalue(data+1, 3);

  fprintf(xml, "<object class=\"%02X\" subclass=\"%06X\" name=\"",
	class, subclass);
  putxmlstr(xml, data+4, iso_to_utf8, 8);
  fputs("\">", xml);
  data += 16;

  class &= 0x7f;

  while (data < end) {
	int code;
	uint32_t len;

	code = *data; data++;
	len = *(uint32_t*) data; data+=4;

	fprintf(xml, "<chunk compression=\"%d\">\n", code);
	switch(code) {
		case 0:
			dumpchunk(xml, data, class);
			break;
		case 1:
			chunk = rledecode(data, len);
			dumpchunk(xml, chunk, class);
			free(chunk);
			break;
		default:
			die("Can't decode chunk type %d", code);
	}

	fputs("</chunk>", xml);
	data += len;
  }
  fputs("</object>\n", xml);

  fclose(xml);

  free(data);

  return 1;
}

int encode(char *xmlname)
{
  uint32_t ofs, rlelen, chksum;
  struct stat st;

  FILE *xml, *dat;
  char *datname;
  uint8_t *data, *rle;

  datname = my_alloc(strlen(filebase)+5, "datname");
  strcpy(datname, filebase);
  strcat(datname, ".dax");

  xml = fopen(xmlname, "rb");	// rt for real xml!!
  if (!xml) die("Can't read %s: %s", xmlname, strerror(errno));

  if (fstat(fileno(xml), &st)) die("Can't stat %s: %s", xmlname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  data = my_alloc(st.st_size, "data");
  fread(data, 1, st.st_size, xml);
  fclose(xml);

  if (data[3] != 0x11) die("File %s is not a .dat file", datname);

  dat = fopen(datname, "w+b");
  fwrite(data, 1, 12, dat);
  chksum = checksum(data, st.st_size);
  fwrite(&chksum, 4, 1, dat);

  rleencode(data+16, st.st_size-16, &rle, &rlelen);

  data[0] = 1;		// chunk type rle encoded
  fwrite(data, 1, 1, dat);

  fwrite(&rlelen, 4, 1, dat);
  fwrite(rle, 1, rlelen, dat);

  fclose(dat);

  free(rle);
  free(data);

  return 1;
}

int main(int argc, char **argv)
{
  char *fname;

  printf("LocoTool V0.03 - Copyright (C) 2004 by Josef Drexler\n");

#ifdef USEICONV
  iso_to_utf8 = iconv_open("utf-8", "iso-8859-1");
  if (iso_to_utf8 == (iconv_t)-1)
	die("Cannot initialize conversion %s to %s: %s", "iso-8859-1", "utf-8", strerror(errno));

  utf16_to_utf8 = iconv_open("utf-8", "ucs-2");
  if (utf16_to_utf8 == (iconv_t)-1)
	die("Cannot initialize conversion %s to %s: %s", "utf-16", "utf-8", strerror(errno));
#endif

  if (argc < 2) die("Give filename as parameter");

  fname = argv[1];
  filebase = my_alloc(strlen(fname)+1, "filebase");

  strcpy(filebase, fname);
  if (strrchr(filebase, '.'))
	strrchr(filebase, '.')[0] = 0;

  if (stricmp(fname+strlen(fname)-4, ".dmp") == 0) {
	encode(fname);
  } else {
	decode(fname);
  }

  free(filebase);

#ifdef USEICONV
  iconv_close(iso_to_utf8);
  iconv_close(utf16_to_utf8);
#endif

  return 0;
}
